/**
 *	Written 2010 by Tyler Montbriand.  Freeware.
 *	http://burningsmell.org/sensor8/
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tcgets.h"
#include "sensor8.h"
#include "tsm_getopt.h"
#include "mydebug.h"

struct
{
	const char *dev;	// What dev to open, i.e. /dev/ttyUSB0
	const char *mode;	// What serial mode, probably 115200N81
	int verbose;	// Prints verbose debugging output when enabled.
	int usage;	// Prints usage when enabled.
	int raw;	// Prints raw bytes when enabled.
	int lock;	// Uses POSIX advisor locking when enabled.

	/**
	 *	If a sensor is present, its first read will produce ERR.
	 *	Subsequent reads to it, for a short time, might produce
	 *	a spurious temperature like 85 degrees C.
	 *	If the first read fails, it will try again at most
	 *	'trymax' number of times, waiting 'tryms' milliseconds
	 *	between each attempted read.
	 *
	 *	Anything below 500 milliseconds doesn't seem reliable.
	 *	Once they start producing sane numbers they seem to
	 *	keep producing sane numbers however.
	 */
	int tryms;
	/**
	 *	How many times to try to read a valid temp.
	 */
	int trymax;	// Max # of temp read retries
	int rdmin;	// Min # of valid temp reads
} globals={
	.dev	=	"/dev/ttyUSB0",
	.mode	=	"115200N81H",	// Buggy driver might need hardware flow enabled
	.verbose=	0,
	.usage	=	0,
	.raw	=	0,
	.tryms	=	650,
	.trymax	=	1,
	.rdmin	=	3,
	.lock	=	0,
	};

const struct opt mopts[];

int open_dev(void), close_dev(int fd);

static int compare_s(const temp_t *a, const temp_t *b);

typedef int (*compare_func)(const void *a, const void *b);

#define COMMAND(TYPE, REPLY, VAL)	do { cmdbuf[cmdpos]=(TYPE); reply[cmdpos++]=(REPLY)|(VAL); } while(0)
//#define REPLY(BYTES, TYPE, VALUE)	reply[replypos++]=((BYTES<<24)|(TYPE)|(VALUE))
#define REPLYTYPE(DATA)			((DATA)&0x00ff0000)
#define REPLYBYTE(DATA, N)		((DATA>>(8*N))&(0x000000ff))

enum replytype_e
{
	REPLY_NONE	=	0x00000000,
	REPLY_EXACT	=	0x00010000,
	REPLY_VOLT	=	0x00020000,
	REPLY_BIT	=	0x00030000,
	REPLY_IGN	=	0x00040000,
	REPLY_TEMP	=	0x00050000,
};

/**
 *	Program to send commands to a DLP-I08 USB I/O module.
 */
int main(int argc, char *argv[])
{
	temp_t t;
	int fd=-1, arg;

	unsigned char cmdbuf[256], cmdpos=0;
	unsigned int reply[256];
	int n, ret=0, retry=globals.trymax, err, dowait;

	struct {
		unsigned char c;
		int reply;
		short *templist;
		int readings;
	} cbuf[256];

	memset(cbuf, 0, sizeof(cbuf));

	if(getenv("SENSOR8"))
		globals.dev=getenv("SENSOR8");

	arg=tsm_getopt(argc, argv, mopts);
	if((arg < 0) || globals.usage || (globals.rdmin < 1) ||
		(globals.rdmin >= 8))
	{
		fprintf(stderr, 
			"sensor8, utility to talk to a DLP-I08 module\n"
			"Written 2010 Tyler Montbriand.  Freeware.\n\n");
		tsm_usage(argc, argv, mopts);
		fprintf(stderr, "\nValid commands:\n"
			"\t[0-7]\t\tex. 7, reads logic value on pin 7\n"
			"\tV[0-7]\t\tex. V3, reads voltage on pin 3\n"
			"\tT[0-7]\t\tex. T5, reads Celsius on pin 5\n"
			"\tF[0-7]\t\tex. F3, reads Fahrenheit on pin 3\n"
			"\t[0-7]:[01]\tex. 1:0, sets logic 0 on pin 1\n"
			"\tZ[0-7]\t\tex. Z2, tristates pin 2\n"
			"\tACK\t\tPings device, waits for reply\n"
			"\tD[0-9]+\t\tex. D100, delays 100 milliseconds\n"
			"\tU[0-9]+\t\tex. U1000, delays 1000 microseconds\n\n"
			"\t%s ACK 1:1 D100 T0 Z1\n"
			"\t\t...will try to ACK the device, set pin 1 to logic high,\n"
			"\t\tsleep 100 milliseconds, read temp from pin 0, then\n"
			"\t\ttristate pin 1.\n\n"
			"\t\tPrepend @ to any command to make it silent.\n",
			argv[0] );
		return(1);
	}

	if(arg == argc)
	{
		fprintf(stderr, "No commands given\n");
		return(2);
	}

	if(globals.verbose)	setenv("DEBUG", "1", 1);


	fd=open_dev();

	if(fd < 0)
	{
		perror(globals.dev);
		return(1);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_WRLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLKW, &l) != 0)
		{
			perror("Couldn't lock");
			close(fd);
			return(1);
		}
	}
#endif

	{
		// Set binary mode so we don't have to parse things like
		// "999.99*C\r\r"
		char c=CMD_SET_BINARY;
		write(fd, &c, 1);
	}


	arg--;

	// Convert arguments into bytes
	while((++arg) < argc)
	{
		if(strcmp(argv[arg], "PING") == 0)	COMMAND(CMD_PING, REPLY_EXACT, REPLY_PING);
		else if(sscanf(argv[arg], "D:%d", &n) == 1)
			{ usleep(n*1000);	continue; }
		else if(sscanf(argv[arg], "U:%d", &n) == 1)
			{ usleep(n);		continue; }
		else if((argv[arg][0] >= '1') && (argv[arg][0] <= '8'))
		{
			int pin=argv[arg][0]-'1';

			if(argv[arg][1] != ':')
				COMMAND(digital_in[pin], REPLY_BIT, 0);
			else for(n=2; argv[arg][n]; n++)
			switch(toupper(argv[arg][n]))
			{
			case 'V':	COMMAND( analog_in[pin],REPLY_VOLT,0);	break;
			case 'I':	COMMAND(digital_in[pin],REPLY_BIT, 0);	break;
			case '0':	COMMAND(    set_lo[pin],REPLY_NONE,0);	break;
			case '1':	COMMAND(    set_hi[pin],REPLY_NONE,0);	break;
			case 'Z':	COMMAND(digital_in[pin],REPLY_IGN, 0);	break;
			case 'T':	case 'F':
				COMMAND(temp_in[pin], REPLY_TEMP,0);
				break;
			default:
				fprintf(stderr, "Unknown command '%c'\n", argv[arg][n]);
				break;
			}
		}
		else	fprintf(stderr, "Bad command '%s'\n", argv[arg]);
	}

	err=1;	// Prime the loop
	dowait=0;

	while(err)
	{
		unsigned char buf[256], pos=0;

		// Assemble the remaining commands
		for(n=0; n<cmdpos; n++) if(cmdbuf[n]) buf[pos++]=cmdbuf[n];

		buf[pos]=0;

		err=0;

		fprintf(stderr, "retrying %s (wait=%d)\n", buf, dowait);
		fprintf(stderr, "%08x,%08x,%08x,%08x\n", reply[0], reply[1], reply[2], reply[3]);

		if(dowait)
		{
			usleep(globals.tryms * 1000);
			dowait=0;
		}


		if(write_n(fd, buf, pos) != pos)
		{
			perror("write error");
			exit(1);
		}

		for(n=0; n<cmdpos; n++)
		if(cmdbuf[n])
		switch(REPLYTYPE(reply[n]))
		{
			unsigned short volt;
			signed short temp;
			unsigned char rbuf;
			ssize_t r;
		default:
			fprintf(stderr, "Eh?  0x%08x\n", reply[n]);
			break;

		case REPLY_EXACT:
			if((r=read(fd, &rbuf, 1)) != 1)
			{
				perror("short read");
				exit(1);
			}

			if(rbuf != REPLYBYTE(reply[n],0))
			{
				fprintf(stderr, "Bad reply 0x%02x\n", (int)rbuf);
				exit(1);
			}

			fprintf(stderr, "0x%02x == 0x%02x\n", (int)rbuf, REPLYBYTE(reply[n],0));
			// FALL THROUGH
		case REPLY_NONE:
			cmdbuf[n]=0;
			break;

		case REPLY_BIT:
			if(r=read(fd, &rbuf, 1) != 1) goto CMD_ERR;

			fprintf(stderr, "bit=%d\n", (int)rbuf);
			cmdbuf[n]=0;
			break;

		case REPLY_VOLT:
			if(r=read(fd, &volt, 2) != 2)	goto CMD_ERR;

			fprintf(stderr, "v=%.02f\n", (htons(volt)*5)/1024.0);
			cmdbuf[n]=0;
			break;

		case REPLY_IGN:	// We get bytes but don't care what they are
			if(r=read(fd, &rbuf, 1) != 1)	goto CMD_ERR;

			fprintf(stderr, "Ignored 1 byte(%d)\n", n);
			cmdbuf[n]=0;
			break;

		case REPLY_TEMP:
			if(r=read(fd, &temp, 2) != 2)	goto CMD_ERR;

			if(((unsigned short)temp == TEMP_ERROR)||
				((unsigned short)temp==TEMP_RESET))
			{
				dowait=1000;
				fprintf(stderr, "temperature retry\n");
				goto CMD_ERR;
			}

			fprintf(stderr, "0x%04x = %.2fC\n", (unsigned short)temp, temp/16.0f);
			cmdbuf[n]=0;
			break;
		CMD_ERR:
			err++;
			break;
		}
	}


#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_UNLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLK, &l) != 0)
		{
			perror("Couldn't clear lock");
			close_dev(fd);
			return(1);
		}
	}
#endif

	close_dev(fd);
	return(0);
}

int close_dev(int fd)
{
	if(fd >= 0)
	{
		if(strcmp(globals.mode, "-") != 0)	// Change no TTY settings
		if(termios_restore(fd) < 0)
		fprintf(stderr, "Couldn't restore termios\n");

		close(fd);
	}
}

int open_dev(void)
{
	int fd=-1;

	fd=open(globals.dev, O_RDWR);

	if(fd < 0)
		return(-1);

	if(strcmp(globals.mode, "-") == 0)	// Change no TTY settings
		return(fd);

	if(termios_save(fd) < 0)
	{
		close(fd);
		return(-1);
	}

	if(termios_set(fd, globals.mode) < 0)
	{
		close(fd);
		return(-1);
	}

	return(fd);
}


static int compare_s(const temp_t *a, const temp_t *b)
{
	return((*a)-(*b));
}

const struct opt mopts[]={
	{"device",	'd',	1, &(globals.dev), NULL, "-d /dev/ttyUSB0" },
	{"mode",	'm',	1, &(globals.mode),NULL, "-m 115200N81 (- means leave alone)" },
	{"verbose",	'v',	0, &(globals.verbose), NULL, "Enables debug output" },
	{"help",	'h',	0, &(globals.usage), NULL, "Shows this listing" },
	{"raw",		'r',	0, &(globals.raw), NULL, "Outputs raw bytes"},
	{"tempms",	'm',	1, &(globals.tryms), "%d", 
			"Milliseconds between temp read retries" },
	{"retries",	'n',	1, &(globals.trymax), "%d", "Max # of tries to read a temp from a pin"},
	{"count",	'c',	1, &(globals.rdmin), "%d", "Min # of valid temperature reads per sensor, 1-7"},
	{"lock",	'l',	0, &(globals.lock), NULL, "Enables cooperative locking" },
	{NULL,		0,	0,	NULL,	 NULL, 	NULL }
	};
