/**
 *	Written 2010 by Tyler Montbriand.  Freeware.
 *	http://burningsmell.org/sensor8/
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tcgets.h"
#include "sensor8.h"
#include "tsm_getopt.h"
#include "mydebug.h"
#include "i2c.h"
#include "help.h"

struct
{
	const char *dev;	// What dev to open, i.e. /dev/ttyUSB0
	const char *mode;	// What serial mode, probably 115200N81
	int verbose;	// Prints verbose debugging output when enabled.
	int usage;	// Prints usage when enabled.
	int raw;	// Prints raw bytes when enabled.
	int lock;	// Uses POSIX advisor locking when enabled.

	int SCL, SDA;	// Pins for bit-banged I2C

	/**
	 *	If a sensor is present, its first read will produce ERR.
	 *	Subsequent reads to it, for a short time, might produce
	 *	a spurious temperature like 85 degrees C.
	 *	If the first read fails, it will try again at most
	 *	'trymax' number of times, waiting 'tryms' milliseconds
	 *	between each attempted read.
	 *
	 *	Anything below 500 milliseconds doesn't seem reliable.
	 *	Once they start producing sane numbers they seem to
	 *	keep producing sane numbers however.
	 */
	int tryms;
	/**
	 *	How many times to try to read a valid temp.
	 */
	int trymax;	// Max # of temp read retries
	int rdmin;	// Min # of valid temp reads
	int bug;	// workaround the kernel flow-control bug
} globals={
	.dev	=	"/dev/ttyUSB0",
	.mode	=	"115200N81H",	// Buggy driver might need hardware flow enabled
	.verbose=	0,
	.usage	=	0,
	.raw	=	0,
	.tryms	=	650,
	.trymax	=	5,
	.rdmin	=	3,
	.lock	=	0,
	.bug	=	0,
	};

const struct opt mopts[];

int open_dev(void), close_dev(int fd);

static int compare_s(const temp_t *a, const temp_t *b);

typedef int (*compare_func)(const void *a, const void *b);

enum replytype_e
{
	REPLY_NONE	=	0x0000,
	REPLY_VOLT	=	0x0100,
	REPLY_BIT	=	0x0200,
	REPLY_IGN	=	0x0300,
	REPLY_TEMP	=	0x0400,
	REPLY_F		=	0x0500,
};

int transact(int fd, unsigned char c, int reply)
{
	short *templist=NULL;
	union
	{
		unsigned short volt;
		signed short temp;
		unsigned char rbuf;
	} d;
	int size, tnum=0, tries=0;

	switch(reply)
	{
	case 0:			size=0;	break;

	case REPLY_IGN:
	default:		size=1;	break;

	case REPLY_TEMP:
	case REPLY_VOLT:	size=2;	break;	
	}

RETRY:
	if(write(fd, &c, 1) != 1)	{	perror("write error");	return(-1);	}

	if(reply)
	if(read_n(fd, &d, size) != size)	{	perror("short read\n");	return(-1);}

	switch(reply)
	{
		unsigned char rbuf;
		unsigned short volt;
		temp_t temp;
		ssize_t r;
	default:
		if(d.rbuf != reply)	return(-1);
		break;

	case REPLY_NONE:	// No reply
	case REPLY_IGN:		// We get bytes but don't care what they are
		break;

	case REPLY_BIT:
		printf("%d", rbuf);
		break;

	case REPLY_VOLT:
		printf("%.2f\n", (htons(d.volt)*5)/1024.0);
		break;

	case REPLY_F:
	case REPLY_TEMP:
		if(!templist)
			templist=malloc(sizeof(short) * globals.rdmin);
		// These read wrong sometimes.  So, lots of sanity
		// checking.
		if(	((unsigned short)temp == TEMP_ERROR)	||
			((unsigned short)temp == TEMP_RESET)	||
			(temp > TEMP_MAX)			||
			(temp < TEMP_MIN)	)
		{
			if((++tries) < globals.trymax)
			{
				usleep(globals.tryms * 1000);
				goto RETRY;
			}
			return(-1);
		}

		templist[tnum++]=d.temp;
		if(tnum < globals.rdmin)
		{
			usleep(globals.tryms * 1000);
			goto RETRY;
		}

		qsort(templist, sizeof(temp_t), globals.rdmin, (compare_func)compare_s);

		if(reply==REPLY_TEMP)
			printf("%.2fC\n", templist[globals.rdmin/2] / 16.0f);
		else	printf("%.2fF\n", ((templist[globals.rdmin/2] * 9)/80.0f)+32.0);
		break;
	}

	if(templist) free(templist);

	return(0);
}

/**
 *	Program to send commands to a DLP-I08 USB I/O module.
 */
int main(int argc, char *argv[])
{
	int fd=-1, arg;
	unsigned char cmdpos=0;
	int n, err, dowait;

	struct {
		unsigned char	c;
		int		reply;
		temp_t		*templist;
		int		readings;
		int		tries;
		char		out[16];
	} cbuf[256];

	memset(cbuf, 0, sizeof(cbuf));

	if(getenv("SENSOR8"))
		globals.dev=getenv("SENSOR8");

	if(getenv("SCL"))	sscanf(getenv("SCL"), "%d", &globals.SCL);
	if(getenv("SDA"))	sscanf(getenv("SDA"), "%d", &globals.SDA);

	arg=tsm_getopt(argc, argv, mopts);
	if((arg < 0) || globals.usage || (globals.rdmin < 1) ||
		(globals.rdmin >= 8))
	{
		fprintf(stderr, 
			"sensor8 v1.0.0, utility to talk to a DLP-I08 module\n"
			"Written 2012 Tyler Montbriand.  Freeware.\n\n");
		tsm_usage(argc, argv, mopts);
		fprintf(stderr,"%s\n", helptext);
		return(1);
	}

	if(arg == argc)
	{
		fprintf(stderr, "No commands given\n");
		return(2);
	}

	if(globals.verbose)	setenv("DEBUG", "1", 1);

	if(globals.bug)
	{
		char *old=globals.mode;
		globals.mode="115200N81";
		fd=open_dev();
		close(fd);
		globals.mode="115200N81H";
		fd=open_dev();
		close(fd);
	}

	fd=open_dev();

	if(fd < 0)
	{
		perror(globals.dev);
		return(1);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_WRLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLKW, &l) != 0)
		{
			perror("Couldn't lock");
			close(fd);
			return(1);
		}
	}
#endif

	{
		// Set binary mode so we don't have to parse things like
		// "999.99*C\r\r"
		char c=CMD_SET_BINARY;
		write(fd, &c, 1);
	}


	arg--;

	while((++arg) < argc)
	{
		int err=0;
		int devid=-1, addr=-1, byte=-1;

		if(sscanf(argv[arg], "%02x[%02x]=%02x", &devid, &addr, &byte) == 3)
		{
			unsigned char buf[256], bpos=0;
			char *h=strchr(argv[arg], '=');
			h++;

			while(1)
			{
				sscanf(h, "%02hhx", buf+(bpos++));

				h++;
				if(!(*h)) break;
				h++;
				if(!(*h)) break;
			}

			err=cmd8_i2c_write7_multi(fd,globals.SCL-1,globals.SDA-1,devid,addr,buf,bpos);
		}
		else if(sscanf(argv[arg], "%02x[%02x]:%d", &devid, &addr, &byte) > 1)
			err=cmd8_i2c_read7_multi(fd,globals.SCL-1, globals.SDA-1, devid, addr, byte);
		else if(strcmp(argv[arg], "PING") == 0)
		{
			if(transact(fd, CMD_PING, REPLY_PING)<0)
			{
				fprintf(stderr, "Couldn't contact device\n");
				exit(1);
			}
		}
		else if(sscanf(argv[arg], "D:%d", &n) == 1)	{ usleep(n*1000);	continue; }
		else if(sscanf(argv[arg], "U:%d", &n) == 1)	{ usleep(n);		continue; }
		else if((argv[arg][0] >= '1') && (argv[arg][0] <= '8'))
		{
			int pin=argv[arg][0]-'1';

			if(argv[arg][1] != ':')
				err=transact(fd,digital_in[pin], REPLY_BIT);
			else for(n=2; argv[arg][n]; n++)
			switch(toupper(argv[arg][n]))
			{
			case 'V':	err=transact(fd,analog_in[pin],REPLY_VOLT);break;
			case 'I':	err=transact(fd,digital_in[pin],REPLY_BIT);break;
			case '0':	err=transact(fd,set_lo[pin],REPLY_NONE);   break;
			case '1':	err=transact(fd,set_hi[pin],REPLY_NONE);   break;
			case 'Z':	err=transact(fd,digital_in[pin],REPLY_IGN);break;
			case 'T':	err=transact(fd,temp_in[pin], REPLY_TEMP); break;
			case 'F':	err=transact(fd,temp_in[pin], REPLY_F);    break;
			default:
				fprintf(stderr, "Unknown command '%c'\n", argv[arg][n]);
				break;
			}
		}
		else	fprintf(stderr, "Bad command '%s'\n", argv[arg]);

		if(err)	printf("ERR\n");
	}


#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_UNLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLK, &l) != 0)
		{
			perror("Couldn't clear lock");
			close_dev(fd);
			return(1);
		}
	}
#endif

	close_dev(fd);
	return(0);
}

int close_dev(int fd)
{
	if(fd >= 0)
	{
		if(strcmp(globals.mode, "-") != 0)	// Change no TTY settings
		if(termios_restore(fd) < 0)
		fprintf(stderr, "Couldn't restore termios\n");

		close(fd);
	}
}

int open_dev(void)
{
	int fd=-1;

	fd=open(globals.dev, O_RDWR);

	if(fd < 0)
		return(-1);

	if(strcmp(globals.mode, "-") == 0)	// Change no TTY settings
		return(fd);

	if(termios_save(fd) < 0)
	{
		close(fd);
		return(-1);
	}

	if(termios_set(fd, globals.mode) < 0)
	{
		close(fd);
		return(-1);
	}

	return(fd);
}


static int compare_s(const temp_t *a, const temp_t *b)
{
	return((*a)-(*b));
}

const struct opt mopts[]={
	{"device",	'd',	1, &(globals.dev), NULL, "-d /dev/ttyUSB0" },
	{"mode",	'm',	1, &(globals.mode),NULL, "-m 115200N81 (- means leave alone)" },
	{"verbose",	'v',	0, &(globals.verbose), NULL, "Enables debug output" },
	{"bug",		0,	0, &(globals.bug), NULL, "Works around kernel bug"},
	{"help",	'h',	0, &(globals.usage), NULL, "Shows this listing" },
	{"raw",		'r',	0, &(globals.raw), NULL, "Outputs raw bytes"},
	{"tempms",	'm',	1, &(globals.tryms), "%d", 
			"Milliseconds between temp read retries" },
	{"retries",	'n',	1, &(globals.trymax), "%d", "Max # of tries to read a temp from a pin"},
	{"count",	'c',	1, &(globals.rdmin), "%d", "Min # of valid temperature reads per sensor, 1-7"},
	{"lock",	'l',	0, &(globals.lock), NULL, "Enables cooperative locking" },
	{"SCL",		0,	1, &(globals.SCL), "%d", "What pin to use for i2c clock" },
	{"SDA",		0,	1, &(globals.SDA), "%d", "What pin to use for i2c data" },
	{NULL,		0,	0,	NULL,	 NULL, 	NULL }
	};
